var convert, file, fs, path;

fs = require('fs');

path = require('path');

convert = require('./convert');

file = exports;

file.extensionOf = function(filename) {
  var parsed;
  parsed = path.parse(filename);
  if (parsed.ext) {
    return parsed.ext;
  } else if (parsed.base[0] === '.') {
    return parsed.base;
  } else if (`.${filename}` in convert.extensions) {
    return `.${filename}`;
  } else {
    return null;
  }
};

file.toFile = function(fold, output, converter = null) {
  var outFormat, result;
  outFormat = file.extensionOf(output);
  if (!outFormat) {
    console.warn(`Could not detect extension of ${output}`);
    return;
  }
  if (converter == null) {
    converter = convert.getConverter('.fold', outFormat);
    if (converter == null) {
      console.warn(`No converter from .fold to ${outFormat}`);
      return;
    }
  }
  result = converter(fold);
  if (typeof result !== 'string') {
    result = convert.toJSON(result);
  }
  return fs.writeFileSync(output, result, 'utf-8');
};

file.fileToFile = function(input, output, options) {
  var converter, inFormat, outFormat, result;
  inFormat = file.extensionOf(input);
  outFormat = file.extensionOf(output);
  if (!inFormat) {
    console.warn(`Could not detect extension of ${input}`);
    return;
  }
  if (!outFormat) {
    console.warn(`Could not detect extension of ${output}`);
    return;
  }
  converter = options.converter;
  if (converter == null) {
    converter = convert.getConverter(inFormat, outFormat);
    if (converter == null) {
      console.warn(`No converter from ${inFormat} to ${outFormat}`);
      return;
    }
  }
  if (outFormat === output || outFormat === `.${output}`) {
    //# just extension => concatenate
    output = path.parse(input);
    output.ext = outFormat;
    output.base = output.name + output.ext;
    output = path.format(output);
  }
  if (input === output) {
    return console.warn(`Attempt to convert ${input} to same filename`);
  } else {
    console.log(input, '->', output);
    result = fs.readFileSync(input, 'utf-8');
    if ((inFormat === '.fold' && '.fold' !== outFormat)) { // avoid double mogrification
      result = file.mogrify(result, options);
    }
    result = converter(result);
    if (outFormat === '.fold') {
      result = file.mogrify(result, options);
    }
    if (typeof result !== 'string') {
      result = convert.toJSON(result);
    }
    return fs.writeFileSync(output, result, 'utf-8');
  }
};

file.mogrify = function(data, options) {
  var error, fold;
  if (!options.flatFold) { // or any options set
    return;
  }
  fold = JSON.parse(data);
  if (options.flatFold) {
    error = convert.flat_folded_geometry(fold);
    console.log(` -- Flat folding error: ${error}`);
    fold.vertices_flatUnfoldCoords = fold.vertices_coords;
    fold.vertices_coords = fold.vertices_flatFoldCoords;
    fold.frame_classes = fold.frame_classes.filter(function(x) {
      return x !== 'creasePattern';
    }).concat('foldedForm');
    delete fold.vertices_flatFoldCoords;
  }
  return convert.toJSON(fold);
};

file.main = function(args = process.argv.slice(2)) {
  var arg, filename, filenames, i, j, len, len1, mode, options, output, results;
  filenames = [];
  output = '.fold'; //# Default behavior: convert to .fold
  options = {
    flatFold: false
  };
  mode = null;
  for (i = 0, len = args.length; i < len; i++) {
    arg = args[i];
    switch (mode) {
      case 'output':
        output = arg;
        mode = null;
        break;
      default:
        switch (arg) {
          case '-o':
          case '--output':
            mode = 'output';
            break;
          case '--flat-fold':
            options.flatFold = true;
            break;
          default:
            filenames.push(arg);
        }
    }
  }
  results = [];
  for (j = 0, len1 = filenames.length; j < len1; j++) {
    filename = filenames[j];
    results.push(file.fileToFile(filename, output, options));
  }
  return results;
};

if (require.main === module) {
  file.main();
}
